/*******************************************************************************
* Interface of the compiler generator library which is included in the scanner
* interface generated with SLADE
*
* JHG sep97 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/
#ifndef CGL_H
#define CGL_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#ifdef Debugger
#include "debug.h"
#endif /* Debugger */

#ifdef WIN32
int					optind;
char				*optarg;
int					getopt(int argc, char *argv[], char *options);
#elif unix
extern char			*optarg;
extern int			optind;
#else
#include <dir.h>
#endif

#define i_STRINGLEN	255   /* maximum length of string             */

typedef enum		{ false= 0, true } boolean;
typedef int			integer;
typedef double		real;
typedef char		string[i_STRINGLEN+1];

extern int			i_linenr;		/* current line number */
extern int			i_nr_errors;	/* number of (syntax) errors */
extern string		i_lexeme;		/* current token representation */
extern char			*i_lexeme_ptr;	/* end of current token representation */
extern char			*i_lexeme_end;	/* end of current token */
extern int			i_ch;			/* current input character*/
extern char			i_input_name[];

typedef enum {
	i_ERROR, 
	i_EACCNOTA, i_EACCONF, i_EADDRESS, i_EBNDCONF, i_EBNDNOTA, i_EDBLDEF,
	i_EDIMCONF, i_EDIMNOTA, i_EDIVBY0, i_EFINVIM, i_EFREAD, i_EFWRITE,
	i_EILLARG, i_EILLREF, i_EINIVIM, i_EINPUTBUF, i_EKINDCONF,
	i_EKINDNOTA, i_ELENTAB, i_EMODBY0, i_ENOADDR, i_ENOFIELD, i_ENOIDINS,
	i_ENOIND, i_ENOKEYW, i_ENOMDOP, i_ENOMEM, i_ENOSCOPE, i_ENOSTATE,
	i_ENOSTORE, i_ENOTFOUND, i_NOTOKEN, i_EOPARGS, i_ERANGE, i_ERDTYPE,
	i_ESTRTAB, i_ETERMEXP, i_ETOOFEW, i_ETOOMANY, i_ETYPECONF, i_ETYPENOTA,
	i_EWRTYPE,
	
	i_WARNING, 
	i_WDANGLE, i_WGRPINS, i_WNOCALL, i_WNOTSET, i_WNOTUSED, i_WTERMDEL,
	i_WTERMINS,

	i_TRACE,
	i_TINP, i_TLST, i_TNERRS, i_TNITAB, i_TNLTAB, i_TNSTAB, i_TPARSTST,
	i_TSCANTST, i_TUSAGE, i_TVERSION, i_TVIM
} i_ReportNumber;

void				i_report(i_ReportNumber rnr, ...);
char				*i_read_line(char *, int, char *);

char				*get_repr(string repr);
char				*convert_to_upper(string repr);
char				*convert_to_lower(string repr);
char				*delete_spaces(string repr);
integer				int_encode(string repr);
char				char_encode(string repr);
boolean				bool_encode(string repr);
real				real_encode(string repr);
char				*string_encode(string repr);
void				i_init_lexeme(void);
void				i_next_character(void);
void				i_reset_input(void);
boolean				i_init(int ac, char **av);
int					i_finish(void);
void				p_ROOT(void);	/* only for debugable parser */

/* Default attribute actions for standard types (used during error recovery) */
void				i_integer(integer *);
void				i_real(real *);
void				i_boolean(boolean *);
void				i_char(char *);
void				i_string(string *);

/* dotty output prototypes */
void i_add_node(char*,char*, ...);
void i_add_leaf(char*,char*, ...);
void i_parent_node(char*, ...);
void i_create_dot_file(char*, int, int, char*, int, char**);
void i_free_mem(void);
void i_init_dotty(int, char**);
#ifdef WIN32
void run_dotty(char *);			/* run dotty and show the created dot file */
#endif

/* dotty output structures */
struct node
{
	char				*label;
	int					nodenr;
	int					is_leaf;
	struct node			*sibling;
	struct node			*child;
	struct node			*parent;
	struct attributes	*syn_attributes;
	struct attributes	*inh_attributes;
}; 

struct attributes
{
	char				*a_name;
	char				*value;
	struct attributes	*next;
};

#endif /* CGL_H */
