#*******************************************************************************
# Gnu assembler macro definitions for the code generator of SLADE
#
# JHG sep97 SLADE Copyright (C) 1997-1998 University of Twente
#******************************************************************************/
	.lcomm _dim,4
	.lcomm _up,4
	.lcomm _xdpl,4
	.lcomm _size,4
	.lcomm _i,4
	.lcomm _n,4
	.lcomm _dpl,4

# ---------------------------------- xcall -----------------------------------
.macro xcall	retaddr, label
	pushl	\retaddr
	call	_ascall
	addl	$4, %esp
	jmp		\label
.endm

# ---------------------------------- descr -----------------------------------
.macro descr	sn, dpl, dim		# void descr(int sn, int dpl, int dim)

	# This macro mimics the VIM instruction DESCR that creates
	# a dynamic array desriptor, because of stack usage it can
	# not be written in C
	movl	$1, _size		# size= 1;
	movl	\dim, _dim
	jmp	2f
1:
	pushl	_size
	pushl	_dim
	pushl	\dpl
	pushl	\sn
	call	_putxlowsiz		# size= putxlowsiz(sn, dpl, dim, size, lo, up);
	addl	$24, %esp
	movl	%eax, _size	
	decl	_dim			# dim--
2:
	movl	_dim, %ebx
	testl	%ebx, %ebx		# dim != 0
	jnz	1b

	pushl	\sn
	call	_getseglen		# offset= getseglen(sn);
	addl	$4, %esp
	pushl	%eax
	pushl	\dpl+1
	pushl	\sn
	call	_putmemr		# putmemr(sn, dpl+1, offset);
	addl	$12, %esp
	pushl	_size
	pushl	\sn
	call	_incrseglen		# incrseglen(sn, size);
	addl	$8, %esp
	pushl	\dim
	pushl	\dpl
	pushl	\sn
	call	_putmemr		# putmemr(sn, dpl, \dim);
	addl	$12, %esp
.endm

# ---------------------------------- eqn -----------------------------------
.macro eqn	len, sn, dpl		# int eqn(int len, sn, dpl)

	# This macro mimics the VIM instruction EQN that compares
	# two (optionally dynamic) arrays on the stack.
	# Because of stack usage it can not be written in C

	movl	\len, %ebx
	testl	%ebx, %ebx		# if (len == 0)
	jnz	1f
	pushl	\dpl+2
	pushl	\sn
	call	_getmemr
	addl	$8, %esp
	movl	%eax, %ebx		# len= getmemr(sn, dpl+2);
1:
	movl	%esp, %esi		# sp1= Sp-1;
	imul	$4, %ebx		# Note: stack grows(!) with steps of 4
	addl	%ebx, %esp		# Sp -= len;
	movl	%esp, %edi		# sp2= Sp;
	addl	%ebx, %esp		# Sp -= len;
2:
	testl	%ebx, %ebx		# len != 0 ?
	jz	3f
	movl	(%esi), %ecx
	cmpl	%ecx, (%edi)		# if (Stack[sp1] == Stack[sp2])
	jne	4f
	addl	$4, %edi		# sp1--
	addl	$4, %esi		# sp2--
	subl	$4, %ebx		# len--
	jmp	2b
3:
	movl	$1, %eax		# equal, set eax to 1
	jmp	5f
4:
	movl	$0, %eax		# not equal, set eax to 0
5:
.endm

# ---------------------------------- ixdpl -----------------------------------
.macro ixdpl	sn, dpl			# int ixdpl(int sn, int dpl)

	# This macro calculates the displacement of an indexed
	# variable, because of stack usage it can not be written in C

	popl	%ebx			# dim= pop();
	movl	$1, _size		# size= 1;
	movl	$0, _xdpl		# xdpl= 0;
	cmpl	$0, %ebx		# if (dim >= 0) {
	jl	2f
1:
	popl	%eax			# idx= pop();
	popl	%edi			# low= pop();
	negl	%edi			# -low
	addl	%edi, %eax		# idx - low
	imull	_size			# (idx - low)*size
	addl	%eax, _xdpl		# xdpl += (idx - low)*size;
	popl	_size			# size= pop();
	decl	%ebx			# dim--;
	testl	%ebx, %ebx		# while (dim)
	jnz	1b
	addl	\dpl, _xdpl		# xdpl += dpl
	jmp	5f
2:
	pushl	\dpl
	pushl	\sn
	call	_getmemr
	addl	$8, %esp
	movl	%eax, %ebx		# dim= getmemr(sn, dpl);
	jmp	4f
3:
	movl	%ebx, _dim
	pushl	%ebx
	pushl	\dpl
	pushl	\sn
	call	_getxlow
	addl	$12, %esp
	movl	%eax, %edi		# low= getxlow(sn, dpl, dim);
	popl	%eax			# idx= pop();
	negl	%edi			# -low
	addl	%edi, %eax		# idx - low
	imull	_size			# (idx - low)*size
	addl	%eax, _xdpl		# xdpl += (idx - low)*size;
	pushl	_dim
	pushl	\dpl
	pushl	\sn
	call	_getxsize
	addl	$12, %esp
	movl	%eax, _size		# size= getxsize(sn, dpl, dim);
	movl	_dim, %ebx
	decl	%ebx
4:
	testl	%ebx, %ebx		# dim > 0
	jnz	3b
	pushl	\dpl+1
	pushl	\sn
	call	_getmemr		# getmemr(sn, dpl+1)
	addl	$8, %esp
	addl	%eax, _xdpl		# xdpl += getmemr(sn, dpl+1);
5:
	pushl	_xdpl
.endm

# ---------------------------------- ldnvar -----------------------------------
.macro ldnvar	sn, dpl

	# This macro loads an array from the data segment on the stack.
	# First the size of the array is popped off the stack, if the
	# size is zero it is loaded from segment sn at displacement dpl.
	# Because of stack usage it can not be written in C.

	movl	\dpl, _dpl
	popl	_n
	movl	_n, %eax
	testl	%eax, %eax		# if (n == 0)
	jnz	1f
	pushl 	\dpl+2
	pushl 	\sn
	call	_getmemr		# n= getmemr(sn, dpl+2);
	addl	$8, %esp
	movl	%eax, _n
	pushl 	\dpl+1
	pushl 	\sn
	call	_getmemr		# dpl= getmemr(sn, dpl+1);
	addl	$8, %esp
	movl	%eax, _dpl
	movl	_n, %eax
1:
	movl	$0, _i			# i= 0;
2:
	cmpl	_i, %eax		# i < n
	je	3f
	pushl	_dpl
	movl	_i, %eax
	addl	%eax, (%esp)
	pushl	\sn
	call	_getmemr
	addl	$8, %esp
	pushl	%eax			# push(getmemr(sn, dpl+i));
	incl	_i			# i++;
	movl	_n, %eax
	jmp	2b
3:
.endm

# ---------------------------------- stnvar -----------------------------------
.macro stnvar	sn, dpl

	# This macro pops an array off the stack into the data segment.
	# First the size of the array is popped off the stack, if the
	# size is zero it is loaded from segment sn at displacement dpl.
	# Because of stack usage it can not be written in C.

	popl	%eax
	movl	\dpl, _dpl
	testl	%eax, %eax		# if (n == 0) {
	jnz	1f
	pushl 	\dpl+2
	pushl 	\sn
	call	_getmemr		#	n= getmemr(sn, dpl+2);
	addl	$8, %esp
	movl	%eax, _n
	pushl 	\dpl+1
	pushl 	\sn
	call	_getmemr		#	dpl= getmemr(sn, dpl+1);
	addl	$8, %esp
	movl	%eax, _dpl		# }
	movl	_n, %eax
1:
	testl	%eax, %eax		# while (n) {
	jz	2f
	decl	%eax			# 	n--
	movl	%eax, _n
	pushl	_dpl
	movl	_n, %eax
	addl	%eax, (%esp)
	pushl	\sn
	call	_putmemr
	addl	$12, %esp		# 	putmemr(sn, dpl+n, pop());
	movl	_n, %eax
	jmp	1b			# }
2:
.endm

# ---------------------------------- popn -----------------------------------
.macro popn	sn, dpl

	# This macro pops an array off the stack.
	# First the size of the array is popped off the stack, if the
	# size is zero it is loaded from segment sn at displacement dpl.
	# Because of stack usage it can not be written in C.

	popl	%eax
	movl	\dpl, _dpl
	testl	%eax, %eax		# if (n == 0)
	jnz	1f
	pushl 	\dpl+2
	pushl 	\sn
	call	_getmemr		# n= getmemr(sn, dpl+2);
	addl	$8, %esp
	movl	%eax, _n
	pushl 	\dpl+1
	pushl 	\sn
	call	_getmemr		# dpl= getmemr(sn, dpl+1);
	addl	$8, %esp
	movl	%eax, _dpl
	movl	_n, %eax
1:
	testl	%eax, %eax		# while (n)
	jz	2f
	decl	%eax			# n--;
	popl	%ebx			# pop();
	jmp	1b
2:
.endm

