/*******************************************************************************
* Interface file of identifier information and scope administration actions
* which is included in a parser generated with SLADE
*
* JHG sep97 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/
#ifndef IDSCOPE_H
#define IDSCOPE_H

#include "cgl.h"

typedef enum {
	bool_type, char_type, int_type, no_type, string_type, unknown_type
} e_type;

typedef enum {
	avar_kind, const_kind, proc_kind, ptr_kind, rec_kind,
	rpar_kind, svar_kind, tdecl_kind, unknown_kind, vpar_kind
} e_kind;

typedef enum {
	address_access= 1, direct_access= 2, indirect_access= 4, readonly_access= 8,
	unknown_access= 0
} e_access;

typedef struct {
	int	low, size;
} BoundRec;

typedef struct DefRec {
	e_access			access;
	integer				dim;
	e_kind				kind;
	e_type				type;
	char				*name;
	integer				sn, dva;
	BoundRec			*bound;
	boolean				delete;
	boolean				*set_or_use;
	integer				nchildren;
	struct DefRec		*sibling;
	struct DefRec		*child;
} DefRec;
typedef DefRec			*def_ptr;

extern def_ptr			undeclared;	/* empty definition record */
extern def_ptr			null_def;	/* null pointer definition record */

#define newline_char	'\n'
#define dim0			0
#define no_dva			-1


void assign_type(e_type *, e_type);
void assign_kind(e_kind *, e_kind);
void assign_definition(def_ptr *, def_ptr);
void assign_access(e_access *, e_access);

void clear_type(e_type *);
void clear_kind(e_kind *);
void clear_definition(def_ptr *);
void clear_access(e_access *);

void check_access(e_access, e_access, boolean);
void check_dim(integer, integer, boolean);
void check_kind(e_kind, e_kind, boolean);
void check_type(e_type, e_type, boolean);
void check_definition(def_ptr, def_ptr, boolean);

void chk_def_access(e_access, def_ptr, boolean);
void chk_def_dim(integer, def_ptr, boolean);
void chk_def_kind(e_kind, def_ptr, boolean);
void chk_def_type(e_type, def_ptr, boolean);

void pointer_check(def_ptr, def_ptr);

void positive_check(integer, e_kind);
void zero_check(integer, e_kind);
void first_parm(def_ptr *, integer *, def_ptr def_proc);
void next_parm(def_ptr *def);
void arg_check(def_ptr argdef, def_ptr pardef);
void arg_access(e_access *, def_ptr pardef);

void result_access(e_access *, e_access, def_ptr);
void result_definition(def_ptr *, def_ptr, def_ptr);
void result_kind(e_kind *, e_kind, def_ptr);
void result_type(e_type *, e_type, def_ptr);
void if_definition(def_ptr *, def_ptr, def_ptr);
void of_type(def_ptr *, def_ptr, integer);

/* scope administration */

void current_scope(integer *);
void enter_scope(void);
void exit_scope(void);

/* create, insert, find and delete identifier definition record pointers */

void create_definition(def_ptr *, integer, e_type, e_kind, integer);
void add_def_bound(def_ptr, int low, int up);
void add_def_child(def_ptr, def_ptr);
void delete_definition(def_ptr);
void find_definition(def_ptr *, string);
void find_field(def_ptr *, string, def_ptr recdef);
void insert_definition(string, def_ptr);


/* get/put the values from/in the definition record */

void get_def_access(e_access *, def_ptr);
void get_def_dim(integer *, def_ptr);
void get_def_kind(e_kind *, def_ptr);
void get_def_type(e_type *, def_ptr);

void put_def_access(e_access, def_ptr);
void put_def_dim(integer, def_ptr);
void put_def_kind(e_kind, def_ptr);
void put_def_type(e_type, def_ptr);

void list_definition(def_ptr);
void list_scope(void);

void use_definition(def_ptr);

char *i_e_access_text(e_access);
char *i_def_ptr_text(def_ptr);
char *i_e_kind_text(e_kind);
char *i_e_type_text(e_type);

#endif /* IDSCOPE_H */
