(*******************************************************************************
* Interface of identifier information and scope administration actions, to be
* included in a (batch oriented) SLADE parser specification
*
* JHG sep97 Copyright (C) 1991-1997 University of Twente
*******************************************************************************)

TYPES
    def_ptr.
    e_access = (
        address_access, direct_access, indirect_access, readonly_access,
        unknown_access
    ). 
    e_kind = (
        avar_kind, const_kind, proc_kind, ptr_kind, rec_kind,
        rpar_kind, svar_kind, tdecl_kind, unknown_kind, vpar_kind
    ). 
    e_type = (
        bool_type, char_type, int_type, no_type, string_type, unknown_type
    ).

CONSTANTS
    dim0         : integer.
    newline_char : char.
    no_dva       : integer.
    null_def     : def_ptr.
    undeclared   : def_ptr.

ACTIONS
    add_def_bound    (in def: def_ptr;    in low: integer;  in up: integer).
    add_def_child    (in parent: def_ptr; in child: def_ptr).
    arg_access       (out a: e_access;    in formal: def_ptr).
    arg_check        (in argdef: def_ptr; in pardef: def_ptr).
    assign_access    (out dest: e_access; in src: e_access).
    assign_definition(out dest: def_ptr;  in src: def_ptr).
    assign_kind      (out dest: e_kind;   in src: e_kind).
    assign_type      (out dest: e_type;   in src: e_type).
    check_access     (in a1: e_access;    in a2: e_access;  in equal: boolean).
    check_dim        (in dim1: integer;   in dim2: integer; in equal: boolean).
    check_kind       (in k1: e_kind;      in k2: e_kind;    in equal: boolean).
    check_type       (in t1: e_type;      in t2: e_type;    in equal: boolean).
    check_definition (in def1: def_ptr;   in def2: def_ptr; in equal: boolean).
    chk_def_access   (in a: e_access;     in def: def_ptr;  in equal: boolean).
    chk_def_dim      (in dim: integer;    in def: def_ptr;  in equal: boolean).
    chk_def_kind     (in k: e_kind;       in def: def_ptr;  in equal: boolean).
    chk_def_type     (in t: e_type;       in def: def_ptr;  in equal: boolean).
    clear_access     (out dest: e_access).
    clear_definition (out dest: def_ptr).
    clear_kind       (out dest: e_kind).
    clear_type       (out dest: e_type).
    create_definition(out def: def_ptr;   in dva: integer;  in t: e_type;
                                          in k: e_kind;     in dim: integer).
    current_scope    (out level: integer).
    delete_definition(in def: def_ptr).
    enter_scope      ().
    exit_scope       ().
    find_definition  (out def: def_ptr;   in name: string).
    find_field       (out def: def_ptr;
                                          in name: string;  in recdef: def_ptr).
    first_parm       (out parm: def_ptr;  out npars: integer;
                                          in proc: def_ptr).
    get_def_access   (out dest: e_access; in src: def_ptr).
    get_def_dim      (out dpl: integer;   in def: def_ptr).
    get_def_kind     (out dest: e_kind;   in def: def_ptr).
    get_def_type     (out dest: e_type;   in def: def_ptr).
    if_definition    (out ifdef: def_ptr; in def1: def_ptr; in def2: def_ptr).
    insert_definition(in name: string;    in def: def_ptr).
    list_definition  (in def: def_ptr).
    list_scope       ().
    next_parm        (out pardef: def_ptr).
    of_type          (out def: def_ptr;   in typ: def_ptr;  in dpl: integer).
    pointer_check    (in def1: def_ptr;   in def2: def_ptr).
    positive_check   (in nr: integer;     in k: e_kind).
    put_def_access   (in a: e_access;     in def: def_ptr).
    put_def_dim      (in d: integer;      in def: def_ptr).
    put_def_kind     (in k: e_kind;       in def: def_ptr).
    put_def_type     (in t: e_type;       in def: def_ptr).
    result_access    (out ra: e_access;   in a: e_access;   in def: def_ptr).
    result_definition(out result: def_ptr;in def1: def_ptr; in def2: def_ptr).
    result_kind      (out result: e_kind; in k: e_kind;     in def: def_ptr).
    result_type      (out result: e_type; in t: e_type;     in def: def_ptr).
    update_dpl       (out dpl: integer;   in def: def_ptr).
    use_definition   (in def: def_ptr).
    zero_check       (in nr: integer;     in k: e_kind).
