/*******************************************************************************
* Implementation of scanner actions which is included in a scanner generated
* with SLADE
*
* JHG sep97 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/

i_token_set			i_look_ahead;
boolean				i_retract;

#define i_NKEYWORDS	((int)(sizeof i_keywords / sizeof i_keywords[0]) - 1)

static void
i_final_accept(i_token_set token)
{
	i_look_ahead= token;
	i_retract= true;
	i_lexeme_end= 0;
}

static void
i_could_accept(i_token_set token)
{
	if (i_look_ahead != token) {
		i_look_ahead= token;
		i_report(i_TSCANTST, "Scanner could recognize %s"
		, i_token_names[i_look_ahead]);
	}
	i_lexeme_end= i_lexeme_ptr-1;
}

static void
i_get_lexeme(void)
{
	if (i_ch != EOF) {
		if (i_lexeme_end) {
			char	*s;

			for (s= i_lexeme_ptr; s >= i_lexeme_end; s--)
				*(s+1)= *s;
		} else {
			i_lexeme_end= i_lexeme_ptr-1;
			*i_lexeme_ptr= *i_lexeme_end;
		}
		*i_lexeme_end= 0;
		i_lexeme_ptr++;
		i_lexeme_end++;
	}
	if (i_look_ahead < 0)
		i_report(i_NOTOKEN, i_lexeme);
	else
		i_report(i_TSCANTST, "Scanner recognized %s, lexeme = '%s'"
		, i_token_names[i_look_ahead], i_lexeme);
}

/* Keyword handling */

static int
i_binsearch(char *lexeme)
{
	int		l, r, m;

	l= 0;
	r= i_NKEYWORDS;
	while (l < r) {
		m= (l + r) / 2;
		if (strcmp(i_keywords[m].rhs, lexeme) <= 0)
			l= m + 1;
		else
			r= m;
	}
	return l;
}

static boolean
i_lookup(char *lexeme)
{
	int		index;

	if ((index= i_binsearch(lexeme)) < 1)
		return false;

	if (strcmp(lexeme, i_keywords[index - 1].rhs))
		return false;

	i_report(i_TSCANTST, "Scanner found keyword %s (derived from %s)"
	, i_token_names[i_keywords[index-1].lhs], i_token_names[i_look_ahead]);
	i_look_ahead= i_keywords[index-1].lhs;
	return true;
}

boolean
i_in_set(i_token_set look_ahead, ...)
{
	va_list		args;
	i_token_set	token;

	va_start(args, look_ahead);
	while ((token= va_arg(args, i_token_set)) != -1)
		if (look_ahead == token)
			return true;

	va_end(args);
	return false;
}	/* i_in_set */

/* used for error recovery: */
void
i_delete_until(i_marker_set markers, ...)
{
	va_list	args;
	i_token_set	token;

	va_start(args, markers);
	while ((token= va_arg(args, i_token_set)) != -1)
		markers[token]= 1;
	va_end(args);

	if (markers[i_look_ahead] == 0) {
		do {
			i_report(i_WTERMDEL, i_token_names[i_look_ahead], i_lexeme);
			i_next_token();
		} while (markers[i_look_ahead] == 0);
	}
}	/* i_delete_until */

