/*******************************************************************************
* Implementation of the compiler generator library containing actions used by
* the scanner and parser generated with SLADE
*
* JHG sep97 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/
#include "cgl.h"

#if defined(unix) || defined(WIN32)
#define MAXDRIVE	1
#define MAXDIR		255
#define MAXFILE		80
#define MAXEXT		40
#define MAXPATH		(MAXDIR+MAXFILE+MAXEXT)
#endif /* unix || WIN32 */

string				i_lexeme;		/* current token representation */
char				*i_lexeme_ptr;	/* end of current token representation */
char				*i_lexeme_end;	/* end of current token representation */
int					i_ch;			/* current input character*/
integer				i_linenr, i_nr_errors;
char				i_input_name[MAXFILE];

static boolean		AtNewLine, CopyLastLine, LstReset,
					TestScanOpt, TestParsOpt;
static char			InputPath[MAXPATH];
static int			LastLineLen;
static FILE			*InpFp;
static FILE			*LstFp;

static char *Reports[]= {
	/* error reports */ "",										/* i_ERROR    */
	"Access '%s' not allowed here",								/* i_EACCNOTA */
	"Access conflict: %s <-> %s",								/* i_EACCONF  */
	"Cannot change address of %s",								/* i_EADDRESS */
	"Boundary size conflict: %s <-> %s",						/* i_EBNDCONF */
	"Equal boundary sizes %d, not allowed here",				/* i_EBNDNOTA */
	"Double defined identifier: '%s'",							/* i_EDBLDEF  */
	"Dimension conflict: %d <-> %d",							/* i_EDIMCONF */
	"Dimension %d not allowed here",							/* i_EDIMNOTA */
	"Divide by zero",											/* i_EDIVBY0  */
	"VIM code not initialised or already finalised",			/* i_EFINVIM  */
	"File '%s' does not exist or is unreadable",				/* i_EFREAD   */
	"File '%s' not writable, cannot continue",					/* i_EFWRITE  */
	"Illegal argument of emit_jump(): '%s'",					/* i_EILLARG  */
	"Illegal reference argument, %s is read-only",				/* i_EILLREF  */
	"VIM code already initialised",								/* i_EINIVIM  */
	"Scanner input buffer overflow",							/* i_EINPUTBUF*/
	"Kind conflict: %s <-> %s",									/* i_EKINDCONF*/
	"Kind '%s' not allowed here",								/* i_EKINDNOTA*/
	"Length table full (more then %d items)",					/* i_ELENTAB  */
	"Modulo by zero",											/* i_EMODBY0  */
	"%s is read-only, cannot be accessed by address",			/* i_ENOADDR  */
	"No field '%s' in record '%s'",								/* i_ENOFIELD */
	"Insertion of identifier '%s' failed, no current scope",	/* i_ENOIDINS */
	"%s cannot be accessed indirectly",							/* i_ENOIND   */
	"Scanner could not find keyword: '%s'",						/* i_ENOKEYW  */
	"No multi-dimensional version of operation '%s' available",	/* i_ENOMDOP  */
	"Out of memory at line %d in %s, cannot continue",          /* i_ENOMEM   */
	"No current scope, exit_scope() failed",					/* i_ENOSCOPE */
	"Scanner has no initial state for input character '%c'",	/* i_ENOSTATE */
	"Cannot store, %s is read-only",							/* i_ENOSTORE */
	"Identifier '%s' not found",								/* i_ENOTFOUND*/
	"Scanner skipped unknown token, lexeme = '%s'",				/* i_ENOTOKEN */
	"Cannot emit operation '%s' without arguments",				/* i_EOPARGS  */
	"Lower bound %d exceeds upper bound %d",					/* i_ERANGE   */
	"No VIM read instruction for type '%s'",					/* i_ERDTYPE  */
	"String table full (more then %d items)",					/* i_ESTRTAB  */
	"Parser expected (non)terminal '%s'",						/* i_ETERMEXP */
	"Too few %s",												/* i_ETOOFEW  */
	"Too many %s",												/* i_ETOOMANY */
	"Type conflict: %s <-> %s",									/* i_ETYPECONF*/
	"Type '%s' not allowed here",								/* i_ETYPENOTA*/
	"No VIM write instruction for type '%s'",					/* i_EWRTYPE  */

	/* warning reports */ "",									/* i_WARNING  */
	"Possible dangling pointer reference",						/* i_WDANGLE  */
	"Parser inserted unnamed syntactical unit (group)",			/* i_WGRPINS  */
	"Procedure '%s' defined but never called",					/* i_WNOCALL  */
	"Variable '%s' might be used before set",					/* i_WNOTSET  */
	"Identifier '%s' defined but never set or used",			/* i_WNOTUSED */
	"Parser deleted terminal '%s' (lexeme = '%s')",				/* i_WTERMDEL */
	"Parser inserted (non)terminal '%s'",						/* i_WTERMINS */

	/* trace reports */ "",										/* i_TRACE    */
	"- input file     : %s",									/* i_TINP	*/
	"- listing file   : %s",									/* i_TLST     */
	"- errors         : %d error(s) detected",					/* i_TNERRS   */
	"%18d VIM instruction(s)",									/* i_TNITAB   */
	"%18d length table item(s)",								/* i_TNLTAB   */
	"%18d string table item(s)",								/* i_TNSTAB   */
	"",															/* i_TPARSTST */
	"",															/* i_TSCANTST */
	"Usage: %s [-ts] [-tp] [-h] file\n"
	"-ts:   test scanner\n"
	"-tp:   test parser\n"
	"-h:    print this information",							/* i_TUSAGE   */
	"Parser %s generated by SLADE v2.9",						/* i_TVERSION */
	"- vim code       : %s",									/* i_TVIM	*/
};

void
i_report(i_ReportNumber rnr, ...)
{
	va_list	args;
	char	buf[2*i_STRINGLEN+1];
	char	*report, *bufp;

	va_start(args, rnr);
	if ((rnr == i_TPARSTST && !TestParsOpt)
	|| (rnr == i_TSCANTST && !TestScanOpt))
		return;


	if (rnr != i_TRACE && rnr != i_TPARSTST && rnr != i_TSCANTST) {
		if (rnr == i_ERROR || rnr == i_WARNING)
			report= va_arg(args, char *);
		else
			report= Reports[rnr];
		buf[0]= 0;
		if (rnr < i_WARNING) {								/* error report */
			strcpy(buf, "Error: ");
			i_nr_errors++;
		} else if (rnr < i_TRACE)							/* warning report */
			strcpy(buf, "Warning: ");

		bufp= buf + strlen(buf);;
	} else {
		report= va_arg(args, char *);
		bufp= buf;
	}
	vsprintf(bufp, report, args);
	va_end(args);

	if (rnr == i_TNERRS && i_nr_errors > 0)
		strcat(buf, ", cannot continue.");

	if (LstFp != 0) {
		if (!LstReset) {
			fseek(LstFp, (long)(i_lexeme_end - i_lexeme_ptr), SEEK_CUR);
			LstReset= true;
		}
		if (rnr < i_TRACE) {
			fprintf(LstFp, "\n*** %s", buf);
			CopyLastLine= true;
		} else if ((rnr == i_TPARSTST && TestParsOpt)
		|| (rnr == i_TSCANTST && TestScanOpt)) {
			fprintf(LstFp, "\n%s", buf);
			CopyLastLine= true;
			return;
		}
	}

#ifdef Debugger
	if (rnr >= i_TRACE) {
		char	*s, *bufp;

		bufp= buf;
		if ((s= strchr(bufp, '\n')) != 0) {
			do {
				*s= 0;
				i_add_list_item(IdOutput, bufp);
				bufp= s+1;
			} while ((s= strchr(bufp, '\n')) != 0);
		}
		if (*bufp)
			i_add_list_item(IdOutput, bufp);
	} else
		i_add_list_item(-IdOutput, buf);
#else
	if (i_linenr > 0 && rnr < i_TRACE)
		fprintf(stderr, "%s:%d: ", InputPath, i_linenr);

	if (rnr != i_TRACE) {
		strcat(buf, "\n");
		fputs(buf, stderr);
	} else 
		fputs(buf, stdout);

	if (rnr == i_ENOMEM)
		exit(1);
#endif /* Debugger */
}

/* Default actions for the standard types */

void
i_integer(integer *integer_value)
{
	*integer_value= 0;
}

void
i_real(real *real_value)
{
	*real_value= 0.0;
}

void
i_boolean(boolean *boolean_value)
{
	*boolean_value= false;
}

void
i_char(char *char_value)
{
	*char_value= 0;
}

void
i_string(string *string_value)
{
	(*string_value)[0]= 0;
}

char *
i_read_line(char *s, int len, char *prompt)
{
#ifdef Debugger
	return i_read_line_dialog(s, len, prompt);
#else
	if (prompt && *prompt)
		fputs(prompt, stdout);
	if (fgets(s, len, stdin))
		s[strlen(s)-1]= 0;	/* remove trailing newline */
	return s;
#endif /* Debugger */
}

#ifdef Debugger
void
i_reset_cgl_files(void)
{
	if (LstFp != 0) {
		fclose(LstFp);
		LstFp= 0;
	}

	if (InpFp != 0) {
		fclose(InpFp);
		InpFp= 0;
	}
}
#endif /* Debugger */

/* Standard scanner actions */

char *
get_repr(string repr)
{
	return repr;
}

char *
convert_to_upper(string repr)
{
	static string	result;
	char			*s;

	strcpy(result, repr);
	for (s= result; *s; s++) 
		if (islower(*s))
			*s= toupper(*s);

	return result;
}

char *
convert_to_lower(string repr)
{
	static string	result;
	char			*s;

	strcpy(result, repr);
	for (s= result; *s; s++) 
		if (isupper(*s))
			*s= tolower(*s);

	return result;
}

char *
delete_spaces(string repr)
{
	static string	result;
	char			*s= result;
	int				i, start, leng;

	*s= 0;
	leng= strlen(repr);

	i= 0;
	while (i < leng) {
		start= i;
		while (i < leng && repr[i] != ' ')
			i++;

		strncat(s, repr+start, i-start);

		while (i < leng && repr[i] == ' ')
			i++;
	}
	return result;
}

integer int_encode(string repr)
{
	return atoi(repr);
}

char char_encode(string repr)
{
	if (strlen(repr) >= 3) {
		if (repr[1] == '\\') {
			switch (repr[2]) {
			case 'a': return '\a';
			case 'b': return '\b';
			case 'f': return '\f';
			case 'n': return '\n';
			case 'r': return '\r';
			case 't': return '\t';
			case 'v': return '\v';
			default: return repr[2];
			}
		} else
			return repr[1];
	}
	return repr[0];
}

boolean
bool_encode(string repr)
{
	if (repr[0] == 'T' || repr[0] == 't')
		return true;

	return false;
}

real
real_encode(string repr)
{
	return atof(repr);
}

char *
string_encode(string repr)
{
	static string	result;

	/* remove first and last character (e.g. quotes) */
	strcpy(result, repr+1);
	result[strlen(result) - 1]= 0;
	return result;
}

void
i_init_lexeme(void)
{
	if (i_ch == EOF)
		return;

	if (i_lexeme_end) {
		if (!LstReset && LstFp) {
			fseek(LstFp, (long)(i_lexeme_end - i_lexeme_ptr), SEEK_CUR);
			LstReset= true;
		}
		i_reset_input();
	}
	i_lexeme_ptr= i_lexeme;
	i_lexeme_end= 0;
}

void
i_reset_input(void)
{
	if (InpFp == 0)
		return;
	
	while (i_lexeme_ptr != i_lexeme_end) {
		ungetc(*(--i_lexeme_ptr), InpFp);
		if (*i_lexeme_ptr == '\n')
			AtNewLine= false;
		else
			LastLineLen--;
	}
}

void
i_next_character(void)
{
	static string	lastline;

	if (i_ch == EOF || InpFp == 0 || LstFp == 0)
		return;

	if ((i_ch= fgetc(InpFp)) != EOF) {
		*i_lexeme_ptr++= i_ch;
		if (CopyLastLine)
			fputc('\n', LstFp);

		if (AtNewLine) {
			fprintf(LstFp, "%3d ", ++i_linenr);
			sprintf(lastline, "%3d ", i_linenr);
			LastLineLen= strlen(lastline);
			AtNewLine= false;
		} else if (CopyLastLine) {
			lastline[LastLineLen]= 0;
			fprintf(LstFp, "\n%s", lastline);
		}

		CopyLastLine= false;
		fputc(i_ch, LstFp);
		LstReset= false;
		if (i_ch != '\n') {
			if (LastLineLen < i_STRINGLEN)
				lastline[LastLineLen++]= i_ch;
		} else
			AtNewLine= true;
	} else
		*i_lexeme_ptr++= 0;

	if (i_lexeme_ptr - i_lexeme >= i_STRINGLEN) {
		i_report(i_EINPUTBUF);
		*i_lexeme_ptr= 0;
		i_lexeme_ptr= i_lexeme;
	}
}

int
i_finish(void)
{
	if (LstFp != 0) {
		fputc('\n', LstFp);
		fclose(LstFp);
		LstFp= 0;
	}
	i_report(i_TNERRS, i_nr_errors);
	return i_nr_errors;
}

#if defined(unix) || defined(WIN32)
static void
fnsplit(char *path, char *drive, char *dir, char *fname, char *ext)
{
	char	*sep, *s;

	if (drive)
		*drive= 0;

#if defined(unix)
	if ((sep= strrchr(path, '/')) != 0) {
#else /* WIN32 */
	if ((sep= strrchr(path, '\\')) != 0) {
#endif
		if (dir) {
			int		len= sep - path + 1;
			strncpy(dir, path, len);
			dir[len]= 0;
		}

		if (fname)
			strcpy(fname, sep+1);
	} else {
		if (dir)
			*dir= 0;
		if (fname)
			strcpy(fname, path);
	}

	if (!fname)
		s= path;
	else
		s= fname;

	if ((sep= strrchr(s, '.')) != 0) {
		if (fname)
			*sep= 0;

		if (ext)
			strcpy(ext, sep+1);
	} else if (ext)
		*ext= 0;
}

#endif /* unix || WIN32 */

#ifdef unix
#include <unistd.h>
#else /* msdos or mswindows */

#include <dos.h>

int				optind= 1;		/* index of which argument is next	*/
char			*optarg;		/* pointer to argument of current option */
 
static char		*letP= 0;	/* remember next option char's location */
static char		SW= 0;			/* DOS switch character, either '-' or '/' */

int
getopt(int argc, char *argv[], char *options)
{
	unsigned char ch;
	char *optP;
  
#ifdef WIN32
	SW= '-';
#else /* WIN16 or DOS */
	if (SW == 0) {
		/* get SW using dos call 0x37 */
		_AX= 0x3700;
		geninterrupt(0x21);
		SW= _DL;
	}
#endif
 
	if (argc > optind) {
		if (letP == 0) {
			if ((letP= argv[optind]) == 0 || (*letP != SW && *letP != '-'))
				goto gopEOF;
			letP++;
			if (*letP == SW || *letP == '-') {
				optind++;
				goto gopEOF;
			}
		}
		if (0 == (ch= *(letP++))) {
			optind++;  goto gopEOF;
		}
		if (':' == ch  ||  (optP= strchr(options, ch)) == 0)  
			goto gopError;
		if (':' == *(++optP)) {
			optind++;
			if (0 == *letP) {
				if (argc <= optind)  goto  gopError;
				letP= argv[optind++];
			}
			optarg= letP;
			letP= 0;
		} else {
			if (0 == *letP) {
				optind++;
				letP= 0;
			}
			optarg= 0;
		}
		return ch;
	}
gopEOF:
	optarg= letP= 0;  
	return EOF;
 
gopError:
	optarg= 0;
	return ('?');
}
#endif

static boolean
read_parameters(int ac, char **av)
{
	int		c;

	optind= 1;
	TestScanOpt= false;
	TestParsOpt= false;
	while ((c= getopt(ac, av, "t:h")) != -1)
		switch (c) {
		case 't':
			if (*optarg == 'p')
				TestParsOpt= true;
			else if (*optarg == 's')
				TestScanOpt= true;
			else
				i_report(i_TUSAGE, av[0]);
			break;

		case 'h':
		case '?':
			i_report(i_TUSAGE, av[0]);
			return false;
		}
	
	if (optind < ac) {
		char	dir[MAXDIR], ext[MAXEXT];

		if ((InpFp= fopen(av[optind], "rt")) == 0) {
			fnsplit(av[optind], 0, dir, i_input_name, ext);
			InputPath[0]= 0;
			if (dir[0])
				strcpy(InputPath, dir);
			strcat(InputPath, i_input_name);
			if ((InpFp= fopen(InputPath, "rt")) == 0) {
				i_report(i_EFREAD, InputPath);
				return false;
			}
		} else {
			fnsplit(av[optind], 0, 0, i_input_name, 0);
			strcpy(InputPath, av[optind]);
		}
		optind++;
	} else {
		i_report(i_TUSAGE, av[0]);
		return false;
	}
	i_report(i_TINP, InputPath);
	return true;
}

boolean
i_init(int ac, char **av)
{
	char fname[MAXFILE+MAXEXT+2];

	i_ch= 0;
	i_lexeme_ptr= 0;
	i_lexeme_end= 0;
	i_linenr= 0;
	i_nr_errors= 0;
	LstReset= false;
	CopyLastLine= false;
	AtNewLine= true;
	i_report(i_TVERSION, av[0]);

	if (!read_parameters(ac, av))
		return false;
	
	sprintf(fname, "%s.lst", i_input_name);
	if ((LstFp= fopen(fname, "w")) == 0) {
		i_report(i_EFWRITE, fname);
		return false;
	}
	i_report(i_TLST, fname);
	return true;
}

#ifdef Debugger
#include <tcl.h>
#include <tk.h>

#ifdef WIN32
#include <process.h>
#endif

static Tcl_Interp *Interp;
#endif /* Debugger */

struct node* CurrentNode=0;

static struct attributes *arg_to_attributes(char *label, va_list args);
static void add_node(char*, struct attributes*);
static void print_tree(FILE*, struct node*, int, int, struct attributes*);
static struct node *find_node(struct node *tree, char *nodename, int nodenamenr);
static void free_trees(struct node*);	
static void init_tree(char*);
static char * capitalize(char *);
static int in_attr_list(struct attributes*, char*);
static int nr_of_attributes(char *);
static char** to_array(char *);


#ifdef WIN32
int getopt(int, char **, char *);
#endif

#ifdef Debugger
int
Tcl_AppInit(Tcl_Interp *interp)
{
	int status;

	status = Tcl_Init(interp);
	printf("Tcl_Init: %s\n", Tcl_GetStringResult(interp));
	if (status != TCL_OK) {
		return TCL_ERROR;
	}

	/* Initialize Tk values. */
	status = Tk_Init(interp);
	printf("Tk_Init: %s\n", Tcl_GetStringResult(interp));

	if (status != TCL_OK) {
		return TCL_ERROR;
	}

	return TCL_OK;
}

static void
i_tcl_eval(char *script)
{
	if (Tcl_Eval(Interp, script) != TCL_OK) {
		printf("Error in script!\n");
		Tcl_Exit(1);
	} else {
	#ifdef SYNTAXTREE_DEBUG
		printf("Evaluated: %s\n", script);
	#endif
	}
}
#endif /* Debugger */

/*
 * void i_add_node(char* label, char* attrs, ...)
 * adds a subtree with label to the current tree
 * add inherited attributes with its value to this tree
 * and set the current tree to the new subtree
 *
 * example function call:
 * i_add_node("label","attr1","val1","attr2","val2",0);
 */
void i_add_node(char* label, char* attrs, ...)
{
return;
	if (!CurrentNode) {
		struct attributes* inh_attributes;
		va_list args;

		va_start(args, attrs);
		inh_attributes = arg_to_attributes(attrs, args);
		va_end(args);

		add_node(label, inh_attributes);
#ifdef Debugger
		Interp = Tcl_CreateInterp();

		if (Tcl_AppInit(Interp) != TCL_OK)
			printf("Tcl_AppInit: niet TCL_OK\n");

#ifdef WIN32
		i_tcl_eval("load ../../arc/w32/bin/tcldot.dll");
#else
		i_tcl_eval("load Tcldot.so");
#endif
		i_tcl_eval("set zoomin 1.15");
		i_tcl_eval("set zoomout [expr {1 / $zoomin}]");
		i_tcl_eval("set depth 1");
		i_tcl_eval("set fileName \"noname\"");

		i_tcl_eval(
			"proc message {m} {\n "
			"	set w .message \n"
			"	catch {destroy $w}\n "
			"	toplevel $w \n"
			"   wm title $w \"Message\" \n"
			"	wm iconname $w \"Message\" \n"
			"	label $w.message -text \"\n$m\n\" \n"
			"	pack $w.message -side top -anchor w \n"
			"	update \n"
			"	after 2000 \"destroy $w\" \n"
			"}"
		);
		i_tcl_eval(
			"proc update_entry {w x y} {\n"
			"	$w.entry delete 0 end\n"
			"	$w.entry insert end [$w.l.list get @$x,$y]\n"
			" }"
		);
		i_tcl_eval(
			"proc loadDirectory {w type} {\n"
			"	if {$type != \"\"} {set type .$type}\n"
			"		$w.d.entry delete 0 end\n"
			"       $w.d.entry insert end [pwd]\n"
			"		$w.d.l.list delete 0 end\n"
			"		if {[pwd] != \"/\"} {\n"
			"			$w.d.l.list insert end \"..\"\n"
			"		}\n"
			"	foreach i [lsort [glob -nocomplain *]] {\n"
			"		if {[file isdirectory $i]} {\n"
			"			$w.d.l.list insert end [file tail $i]\n"
			"		}\n"
			"	}\n"
			"	$w.f.l.list delete 0 end\n"
			"	foreach i [lsort [glob -nocomplain *$type]] {\n"
			"		if {! [file isdirectory $i]} {\n"
			"			$w.f.l.list insert end [file tail $i]\n"
			"		}\n"
			"	}\n"
			"}"
		);

		i_tcl_eval(
			"proc loadDirectory_list {w type x y} {\n"
			"	cd [$w.d.l.list get @$x,$y]\n"
			"	loadDirectory $w $type\n"
			"}"
		);
		i_tcl_eval(
			"proc loadDirectory_entry {w type} {\n"
			"	cd [$w.d.entry get]\n"
			"	loadDirectory $w $type\n"
			"}"
		);
		i_tcl_eval(
			"proc saveFileByName {w name type} {\n"
			"	global modified fileName g\n"
			"	if {[catch {open $name w} f]} {\n"
			"		message \"Unable to open file for write:\n$name;\"\n"
			"	}\n"
			"	$g write $f $type\n"
			"	close $f\n"
			"	if {$w != {}} {destroy $w}\n"
			"	message \"Tree is saved as:$name\"\n"
			"}"
		);
		i_tcl_eval(
			"proc saveFileByName_list {w x y type} {\n"
			"set dirName [$w.d.entry get]\n"
			"if {[catch {cd $dirName}]} {\n"
			"	warning \"No such directory:\n$dirName;\"\n"
			"}\n"
			"if {$dirName == \"/\"} {set dirName \"\"}\n"
			"saveFileByName $w $dirName/[$w.f.l.list get @$x,$y] $type\n"
			"}"
		);
		i_tcl_eval(
			"proc saveFileByName_entry {w type} {\n"
			"	set dirName [$w.d.entry get]\n"
			"	if {[catch {cd $dirName}]} {\n"
			"		warning \"No such directory:\n$dirName; return\"\n"
			"	}\n"
			"	if {$dirName == \"/\"} {set dirName \"\"}\n"
			"		saveFileByName $w $dirName/[$w.f.entry get] $type\n"
			"}"
		);

		i_tcl_eval(
			"proc saveFileAs {} {\n" 
			"	global fileName\n" 
			"	set type ps\n" 
			"	set w .save\n" 
			"   catch {destroy $w }\n" 
			"	toplevel $w \n" 
			"	wm title $w \"Save Tree\"\n" 
			"	wm iconname $w \"Save\"\n" 
			"	frame $w.d\n" 
			"	label $w.d.label -text \"Directory:\"\n" 
			"	frame $w.d.l\n" 
			"	listbox $w.d.l.list -width 30 -height 10 -yscrollcommand \"$w.d.l.scroll set\"\n"
			"	bind $w.d.l.list <Double-1> \"loadDirectory_list $w $type %x %y; break\"\n"
			"	bind $w.d.l.list <1> \"update_entry $w.d %x %y\"\n"
			"	scrollbar $w.d.l.scroll -command \"$w.d.l.list yview\"\n"
			"	pack $w.d.l.list $w.d.l.scroll -side left -fill y -expand 1\n"
			"	frame $w.d.space1 -height 3m -width 20\n"
			"	entry $w.d.entry -width 30\n"
			"	frame $w.d.space2 -height 3m -width 20\n"
			"	button $w.d.cancel -text Cancel -command \"destroy $w\"\n"
			"	bind $w.d.entry <Return> \"loadDirectory_entry $w $type\"\n"
			"	pack $w.d.label $w.d.l $w.d.space1 $w.d.entry $w.d.space2 -side top -anchor w\n"
			"	pack $w.d.cancel -side top\n"
			"	frame $w.space -height 3m -width 3m\n"
			"	frame $w.f\n"
			"	label $w.f.label -text \"File:\"\n"
			"	frame $w.f.l\n"
			"	listbox $w.f.l.list -width 30 -height 10 -yscrollcommand \"$w.f.l.scroll set\"\n"
			"	bind $w.f.l.list <Double-1> \"saveFileByName_list $w %x %y $type; break\"\n"
			"	bind $w.f.l.list <1> \"update_entry $w.f %x %y\"\n"
			"	scrollbar $w.f.l.scroll -command \"$w.f.l.list yview\"\n"
			"	pack $w.f.l.list $w.f.l.scroll -side left -fill y -expand 1\n"
			"	frame $w.f.space1 -height 3m -width 20\n"
			"	entry $w.f.entry -width 30\n"
			"	frame $w.f.space2 -height 3m -width 20\n"
			"	button $w.f.load -text Save -command \"saveFileByName_entry $w $type\"\n"
			"	bind $w.f.entry <Return> \"saveFileByName_entry $w $type; break\"\n"
			"	pack $w.f.label $w.f.l $w.f.space1 $w.f.entry $w.f.space2 -side top -anchor w\n"
			"	pack $w.f.load -side top\n"
			"	pack $w.d $w.space $w.f -side left -fill y -expand true\n"
			"	loadDirectory $w $type\n"
			"	$w.f.entry insert end [file rootname [file tail $fileName]].$type\n"
			"}"
		);

		i_tcl_eval(
			"proc layout {c} {\n"
			"global g\n"
			"$c delete all\n"
			"$g layout\n eval [$g render]\n"
			"$c configure -scrollregion [$c bbox all]\n"
			"update\n"
			"}"
		);
		
		i_tcl_eval(
			"proc zoom {c z} {\n"
			" global depth labels \n" 
			" set depth [expr {$depth * $z}]\n"
			" foreach i [array names labels] {\n"
			"	foreach {text font} $labels($i) {break} \n" 
			"		set fontsize [expr {int([lindex $font 1] * $depth)}]\n"
			"		if {$fontsize} {\n" 
			"			$c itemconfigure $i -text $text -font [lreplace $font 1 1 $fontsize]\n" 
			"		} {\n"
			"			$c itemconfigure $i -text { }\n" 
			"		}\n"
			" }\n"
			" set x [$c canvasx [expr {[winfo pointerx $c] - [winfo rootx $c]}]]\n"
			" set y [$c canvasy [expr {[winfo pointery $c] - [winfo rooty $c]}]]\n"
			" $c scale all $x $y $z $z \n"
			" $c configure -scrollregion [$c bbox all]\n"
			" update\n"
			"}\n"
		);

		i_tcl_eval(
			"proc Window { top } {\n"
			" frame $top \n"
			" set f [frame $top.mbar] \n"
			" menubutton $top.mbar.file -text File -menu $top.mbar.file.menu \n"
			" menubutton $top.mbar.view -text View -menu $top.mbar.view.menu \n"
			" pack $top.mbar.file $top.mbar.view -side left \n"
			" menu $top.mbar.file.menu \n"
			" $top.mbar.file.menu add command -label \"Save\" -command { saveFileAs } \n"
			" $top.mbar.file.menu add command -label \"Exit\" -command { exit } \n"
			" menu $top.mbar.view.menu \n"
			" $top.mbar.view.menu add command -label \"Zoom In\" -command {zoom $c $zoomin} \n"
			" $top.mbar.view.menu add command -label \"Zoom Out\" -command {zoom $c $zoomout} \n"
			" $top.mbar.view.menu add command -label \"reset\" -command {layout $c} \n"
			" pack $f -side top -fill x \n"
			" frame $top.c \n"
			" canvas $top.c.canvas -background white -width 15c -height 10c"
			"	-xscrollcommand [list $top.c.xscroll set] "
			"	-yscrollcommand [list $top.c.yscroll set] \n"
			" scrollbar $top.c.xscroll -orient horizontal -command [list $top.c.canvas xview] \n"
			" scrollbar $top.c.yscroll -orient vertical -command [list $top.c.canvas yview] \n"
			" pack $top.c.canvas -side left -fill both -expand true \n"
			" pack $top.c -side top -fill both -expand true \n"
			" grid $top.c.canvas $top.c.yscroll -sticky news \n"
			" grid $top.c.xscroll -sticky news \n"
			" grid rowconfigure $top.c 0 -weight 1 \n"
			" grid columnconfigure $top.c 0 -weight 1 \n"
			" pack $top -side top -fill both -expand true \n"
			" wm title . \"SLADE Tree viewer\" \n" 
			" wm iconname . \"SLADE Tree viewer\" \n" 
			" return $top.c.canvas \n"
			"}"
		);
		i_tcl_eval("set g [dotnew digraph]");
		i_tcl_eval("$g setnodeattributes -fontsize 10");
		i_tcl_eval("$g setnodeattributes -font Times-Roman");
		i_tcl_eval("set c [Window .c]");
	}

	
	{
		char tmp[300];
		struct attributes *tmpattr = CurrentNode->inh_attributes;
		/* points to the next attribute */

		int attrnum = 0; /* counts which is the next attribute */
		sprintf(tmp, "set node%d [$g addnode node%d label \"", CurrentNode->nodenr, CurrentNode->nodenr);

		while (tmpattr) {
			sprintf(tmp+strlen(tmp), "<iattr%d> %s\\n=%s | ", attrnum, tmpattr->a_name, tmpattr->value);
			attrnum++;
			tmpattr = tmpattr->next;
		}

		sprintf(tmp+strlen(tmp), "<nterm> %s\" shape record color gray]\n", CurrentNode->label);

		i_tcl_eval(tmp);
	}
	if (CurrentNode->parent) {
		char tmp[300];
		sprintf(tmp, "$g addedge $node%d $node%d headport :nterm tailport :nterm color gray\n", CurrentNode->parent->nodenr, CurrentNode->nodenr);
		i_tcl_eval(tmp);

		{
			int attrnumcurrent = 0;
			int attrnumparent;
			struct attributes *tempattr = CurrentNode->inh_attributes;
			struct attributes *tempattrp;
			while (tempattr) {		/* if same attribute is in the parent draw a red line */
				attrnumparent = 0;
				tempattrp = CurrentNode->parent->inh_attributes;
				while (tempattrp) {
					if (!strcmp(tempattr->a_name, tempattrp->a_name)) {
						sprintf(tmp, "$g addedge $node%d $node%d headport :iattr%d tailport :iattr%d color red\n",
						CurrentNode->parent->nodenr, CurrentNode->nodenr, attrnumparent, attrnumcurrent);
						i_tcl_eval(tmp);
						break;
					}
					attrnumparent++;
					tempattrp = tempattrp->next;
				}
				attrnumcurrent++;
				tempattr = tempattr->next;
			}
		}
	}
	i_tcl_eval("layout $c\ngrab current\n");
#else
	}
#endif /* Debugger */
}

/*
 * void i_add_leaf(char *label, char* attrs, ...)
 * adds a leaf (terminal) to the current tree
 * a leaf can have zero or one synthesised attribute(s)
 * that will be displayed under the label when printing
 * the current tree remains the current tree but it will
 * have an extra child
 *
 * this function will also display the new terminal
 */
void i_add_leaf(char* label, char* attrs, ...)
{
return;
	if (CurrentNode) {
		va_list args;

		va_start(args, attrs);
		CurrentNode->syn_attributes = arg_to_attributes(attrs, args);
		va_end(args);

		add_node(label, 0);
		CurrentNode->is_leaf = 1;

#ifdef Debugger
		{
			char tmp[300];
			if (CurrentNode->syn_attributes)
				sprintf(tmp, "set node%d [$g addnode node%d label \"%s\\n%s->%s\" shape ellipse color gray]\n", CurrentNode->nodenr, CurrentNode->nodenr, CurrentNode->label, CurrentNode->syn_attributes->a_name, CurrentNode->syn_attributes->value);
			else
				sprintf(tmp, "set node%d [$g addnode node%d label \"%s\" shape ellipse color gray]\n", CurrentNode->nodenr, CurrentNode->nodenr, CurrentNode->label);

			i_tcl_eval(tmp);
		}

		{
			char tmp[300];
			sprintf(tmp, "$g addedge $node%d $node%d tailport :nterm color gray\n", CurrentNode->parent->nodenr, CurrentNode->nodenr);
			i_tcl_eval(tmp);
		}

		i_tcl_eval("layout $c\ngrab current\n");
#endif /* Debugger */

		if (CurrentNode->parent)
			CurrentNode = CurrentNode->parent;
	}
}

/*
 * static struct attributes *arg_to_attributes(char* attrs, va_list args)
 * puts the attributes that were given in arguments to the add_node() and
 * parent_node() functions in a linked list of attribute structures
 */
static struct attributes *
arg_to_attributes(char *attrs, va_list args)
{
	struct attributes *attribute_list;
	if (attrs) {
		struct attributes* temp_attr;
		char *attribute;
		char *value;

		value = va_arg(args, char*);
		attribute_list = (struct attributes*)malloc(sizeof(struct attributes));
		attribute_list->a_name = malloc(strlen(attrs)+1);
		strcpy(attribute_list->a_name,attrs);
		attribute_list->value = malloc(strlen(value)+1);
		strcpy(attribute_list->value,value);
		temp_attr = attribute_list;

		while ((attribute= va_arg(args, char*))) {
			value = va_arg(args, char*);
			temp_attr->next = (struct attributes*)malloc(sizeof(struct attributes));
			temp_attr= temp_attr->next;
			temp_attr->a_name = malloc(strlen(attrs)+1);
			strcpy(temp_attr->a_name,attribute);
			temp_attr->value = malloc(strlen(value)+1);
			strcpy(temp_attr->value,value);
		}
		temp_attr->next=0;
	} else
		attribute_list = 0;
	return attribute_list;
}

/*
 * static void add_node(char* label, struct attributes* inh)
 * adds a new node with the given label and inherited attributes
 * to the CurrentNode
 */
static void
add_node(char* label, struct attributes* inh)
{
	static int number;
	if (!CurrentNode) {
		init_tree(label);
		CurrentNode->is_leaf=0;
		number = 1;
	} else {
		struct node* new_tree = (struct node*)malloc(sizeof(struct node));
		new_tree->label = malloc(strlen(label)+1);
		strcpy(new_tree->label,label);

		new_tree->sibling = 0;
		new_tree->child = 0;
		new_tree->nodenr = number++;
		new_tree->is_leaf = 0;
		new_tree->parent = CurrentNode;

		if (!CurrentNode->child)
			CurrentNode->child = new_tree;
		else {
			struct node* temp_tree;

			temp_tree = CurrentNode->child;
			while (temp_tree->sibling)
				temp_tree= temp_tree->sibling;
			temp_tree->sibling=new_tree;
		}
		CurrentNode = new_tree;	
	}
	CurrentNode->inh_attributes = inh;
}

/*
 * void i_parent_node(char* attrs, ...)
 * adds synthesised attributes to the current tree,
 * sets the parent of the CurrentNode as the CurrentNode
 * and add the synthesised attributes of the CurrentNode
 * and the children in the Tcl/Tk window
 *
 * example function call:
 * i_parent_node("syn_attr1","value1","syn_attr2","value2",0);
 */
void i_parent_node(char* attrs, ...)
{
return;
	if (CurrentNode) {
		va_list args;

		va_start(args, attrs);
		CurrentNode->syn_attributes = arg_to_attributes(attrs, args);
		va_end(args);

		CurrentNode->is_leaf = 0;

#ifdef Debugger
		{
			char tmp[300];
			if (!CurrentNode->is_leaf) { /* add the attributes to the nodes */
				struct attributes *tmpattr = CurrentNode->inh_attributes;
				/*points to the next attribute */

				int attrnum = 0; /* counts which is the next attribute */
				sprintf(tmp, "set node%d [$g addnode node%d label \"", CurrentNode->nodenr, CurrentNode->nodenr);

				while (tmpattr) {
					sprintf(tmp+strlen(tmp), "<iattr%d> %s\\n=%s | ", attrnum, tmpattr->a_name, tmpattr->value);
					attrnum++;
					tmpattr = tmpattr->next;
				}

				sprintf(tmp+strlen(tmp), "<nterm> %s", CurrentNode->label);
				tmpattr = CurrentNode->syn_attributes;
				attrnum = 0;
				while (tmpattr) {
					sprintf(tmp+strlen(tmp), " | <sattr%d> %s\\n=%s", attrnum, tmpattr->a_name, tmpattr->value);
					attrnum++;
					tmpattr = tmpattr->next;
				}

				sprintf(tmp+strlen(tmp), "\" shape record color gray]\n");
			} else {
				/* the gray line to the leaf can't point to node:nterm anymore because ellipses don't support node ports */
				sprintf(tmp, "set tmp [$g findedge node%d node%d]\n", CurrentNode->parent->nodenr, CurrentNode->nodenr);
				i_tcl_eval(tmp);
				sprintf(tmp, "$tmp setattributes headport :\n");
				i_tcl_eval(tmp);
				if (CurrentNode->syn_attributes)
					sprintf(tmp, "set node%d [$g addnode node%d label \"%s\\n%s->%s\" shape ellipse color gray]\n", CurrentNode->nodenr, CurrentNode->nodenr, CurrentNode->label, CurrentNode->syn_attributes->a_name, CurrentNode->syn_attributes->value);
				else
					sprintf(tmp, "set node%d [$g addnode node%d label \"%s\" shape ellipse color gray]\n", CurrentNode->nodenr, CurrentNode->nodenr, CurrentNode->label);
			}
			i_tcl_eval(tmp);
		}

		if (CurrentNode->child) {
			char tmp[300];
			struct node *childnode = CurrentNode->child;
			while (childnode) {
				struct attributes *tempattrchild = childnode->syn_attributes;
				struct attributes *tempattrparent;
				int attrnumchild = 0;
				int attrnumcurrent;
				while (tempattrchild)
				{ /* if same attribute is in the parent, draw a green line */
					attrnumcurrent = 0;
					tempattrparent = CurrentNode->syn_attributes;
					while (tempattrparent) {
						if (!strcmp(tempattrchild->a_name, tempattrparent->a_name)) {
							sprintf(tmp, "$g addedge $node%d $node%d headport :sattr%d tailport :sattr%d dir back color green\n",
							CurrentNode->nodenr, childnode->nodenr, attrnumcurrent, attrnumchild);
							i_tcl_eval(tmp);
							break;
						}
						attrnumcurrent++;
						tempattrparent = tempattrparent->next;
					}
					attrnumchild++;
					tempattrchild = tempattrchild->next;
				}
				childnode = childnode->sibling;
			}
		}

		if (CurrentNode->parent)
			CurrentNode = CurrentNode->parent;
		else {
			i_tcl_eval("layout $c\n");
			i_tcl_eval(
				"foreach i [$c find all] {\n"
				"	if {[$c type $i] == \"text\"} {\n"
				"		set text [$c itemcget $i -text]\n"
				"		set font [$c itemcget $i -font]\n"
				"		set labels($i) [list $text $font]\n"
				"	}\n"
				"}"
			);
			Tk_MainLoop();
		}
		i_tcl_eval("grab current\n");
#endif /* Debugger */
		if (CurrentNode->parent)
			CurrentNode = CurrentNode->parent;
	}
}

/*
 * void i_create_dot_file() calls print_tree for the CurrentNode
 * and start dotty when specified
 * filename: the name of the file to be created
 * depth: how deep from the chosen non-term will the tree be
 * rankdir: print direction (0=left-right or 1=top-bottom)
 * nodename: start printing at this node
 * nodenamenr: the instance of nodename to start printing at
 * argattrs: the attributes to be printed (0 = all)
 */
void
i_create_dot_file(char* filename, int depth, int rankdir, char *nodename, int nodenamenr, char **argattrs)
{
	if (CurrentNode) {
		FILE* fp;
		struct attributes* attrs;		/* all attributes to show */
		/* get the inherited and synthesized attributes */
		if (!argattrs[0])
			attrs = 0;
		else {
			int tmpcount = 1;
			attrs = (struct attributes *)malloc(sizeof(struct attributes));
			attrs->next = 0;
			attrs->a_name = malloc(strlen(argattrs[0])+1);
			strcpy(attrs->a_name, capitalize(argattrs[0]));
			
			while (argattrs[tmpcount]) {
				struct attributes *tmpattr = (struct attributes*)malloc(sizeof(struct attributes));
				tmpattr->next = attrs;
				attrs = tmpattr;
				tmpattr->a_name = malloc(strlen(argattrs[tmpcount])+1);
				strcpy(tmpattr->a_name, capitalize(argattrs[tmpcount]));
				tmpcount++;
			}
		}

		fp = fopen(filename, "w");
		fprintf(fp, "// Dot-file :%s\n",filename);
		fprintf(fp, "// Automatically generated by Slade\n");
		fprintf(fp, "digraph G {\n");
		if (rankdir) {
			fprintf(fp, "\trankdir=\"TB\";\n");
			fprintf(fp, "\tordering=out;\n");
		} else {
			fprintf(fp, "\trankdir=\"LR\";\n");
		}

		if (nodename) {
			struct node *tree = find_node(CurrentNode, nodename, nodenamenr);
			if (!tree) {
				fprintf(stderr,"Warning: node \"%s\" with occurance %d is not found in the Tree\n", nodename, nodenamenr);
				fprintf(stderr,"Printing entire tree\n");
				tree = CurrentNode;
			}
			print_tree(fp, tree, depth, depth, attrs);
		} else
			print_tree(fp, CurrentNode, depth, depth, attrs);

		fprintf(fp,"}");			/* close the 'digraph' and the file */
		fclose(fp);
	}
}


static void
print_tree(FILE* fp, struct node* the_tree, int depthcnt, int depth, struct attributes *attrs)
{
	if ((!depthcnt)&&(depth))
		return;

	if (the_tree->syn_attributes || the_tree->inh_attributes) {
		struct attributes *tmpattr = the_tree->inh_attributes;
		/*points to the next attribute */
		int attrnum = 0; /* counts which is the next attribute */

		fprintf(fp, "\tnode%d [label=\"", the_tree->nodenr);
		while (tmpattr) {
			if (!attrs || in_attr_list(attrs, capitalize(tmpattr->a_name))) {
				fprintf(fp, "<iattr%d> %s=\\n%s | ", attrnum, tmpattr->a_name, tmpattr->value);
				attrnum++;
			}
			tmpattr = tmpattr->next;
		}
		fprintf(fp, "<nterm> %s", the_tree->label);
		tmpattr = the_tree->syn_attributes;
		attrnum = 0;
		while (tmpattr) {
			if (!attrs || in_attr_list(attrs, capitalize(tmpattr->a_name))) {
				fprintf(fp, " | <sattr%d> %s=\\n%s", attrnum, tmpattr->a_name, tmpattr->value);
				attrnum++;
			}
			tmpattr = tmpattr->next;
		}
		fprintf(fp, "\", shape=record, color=gray];\n");
	} else if (the_tree->child) {
		fprintf(fp, "\tnode%d [label=\"%s\", shape=record,color=gray];\n",the_tree->nodenr,
		the_tree->label);
	} else
		fprintf(fp, "\tnode%d [label=\"%s\", shape=ellipse,color=gray];\n",the_tree->nodenr, the_tree->label);


	if (the_tree->parent && depth != depthcnt) {
		/* draw grey line */
		fprintf(fp, "\tnode%d", the_tree->parent->nodenr);
		if (the_tree->parent->syn_attributes)
			fprintf(fp, ":nterm");
		fprintf(fp, " -> node%d", the_tree->nodenr);
		if (the_tree->syn_attributes)
			fprintf(fp, ":nterm");
		fprintf(fp, " [dir=none, color=gray];\n");
		/* draw red line (inherited attributes) */
		{
			int attrnumcurrent = 0;
			int attrnumparent;
			struct attributes *tempattr = the_tree->inh_attributes;
			struct attributes *tempattrp;
			while (tempattr) { 
				if (!attrs || in_attr_list(attrs, capitalize(tempattr->a_name))) {
					attrnumparent = 0;
					tempattrp = the_tree->parent->inh_attributes;
					while (tempattrp) {
						if (!attrs || in_attr_list(attrs, capitalize(tempattrp->a_name))) {
							if (!strcmp(tempattr->a_name, tempattrp->a_name)) {
								fprintf(fp, "\tnode%d:iattr%d -> node%d:iattr%d[dir=forward,color=red];\n",
								the_tree->parent->nodenr, attrnumparent,the_tree->nodenr,attrnumcurrent);
								break;
							}
							attrnumparent++;
						}
						tempattrp = tempattrp->next;
					}
					attrnumcurrent++;
				}
				tempattr = tempattr->next;
			}
		}
		/* draw green line (synthesized attributes) */
		{
			int attrnumcurrent = 0;
			int attrnumparent;
			struct attributes *tempattr = the_tree->syn_attributes;
			struct attributes *tempattrp;
			while (tempattr) {
				if (!attrs || in_attr_list(attrs, capitalize(tempattr->a_name))) {
					attrnumparent = 0;
					tempattrp = the_tree->parent->syn_attributes;
					while (tempattrp) {
						if (!attrs || in_attr_list(attrs, capitalize(tempattrp->a_name))) {
							if (!strcmp(tempattr->a_name, tempattrp->a_name)) {
								fprintf(fp, "\tnode%d:sattr%d -> node%d:sattr%d[dir=back,color=green];\n",
								the_tree->parent->nodenr, attrnumparent,the_tree->nodenr,attrnumcurrent);
								break;
							}
							attrnumparent++;
						}
						tempattrp = tempattrp->next;
					}
					attrnumcurrent++;
				}
				tempattr = tempattr->next;
			}
		}
	}

	if (the_tree->child) {
		struct node *tmp = the_tree->child;
		while (tmp) {
			print_tree(fp, tmp, depthcnt-1, depth, attrs);
			tmp = tmp->sibling;
		}
	}
}

#ifdef WIN32
/*
 * static void runDotty(char *filename)
 */
static void
i_run_dotty(char *filename)
{
	if (spawnlp(1, "h:\\graphviz\\bin\\dotty.exe", "dotty", filename, '\0') < 0) {
		fprintf(stderr,"Starting dotty failed\n");
		exit(1);
	}
}
#endif

/*
 * static struct node *find_node(struct node *, char *, int)
 * finds the node with the specified label and with the specified instance
 */
static struct node *
find_node(struct node *tree, char *nodename, int nodenamenr)
{
	static int found = 0;	/* Number of 'name' found in the tree */
	static struct node *last_of_nodename;
	if (!strcmp(tree->label, nodename)) {
		found++;
		last_of_nodename = tree;
		if (found == nodenamenr)
			return tree;
	}
	if (tree->child) {
		struct node *tmptree;
		struct node *tmp = tree->child;
		while (tmp) {
			tmptree = find_node(tmp, nodename, nodenamenr);
			if (tmptree)
				return tmptree;
			tmp = tmp->sibling;
		}
	}
	if (!tree->parent && found) {
		fprintf(stderr, "Warning: occurence %d of %s is not found: using %d\n",
			nodenamenr, last_of_nodename->label, found);
		return last_of_nodename;
	} else
		return 0;
}

/*
 * static void init_tree(char* l) init the tree with label l
 */
static void
init_tree(char* l)
{
	CurrentNode = (struct node*)malloc(sizeof(struct node));
	CurrentNode->label = malloc(strlen(l)+1);
	strcpy(CurrentNode->label,l);
	CurrentNode->nodenr = 0;
	CurrentNode->child = 0;
	CurrentNode->parent = 0;
}

/*
 * void i_free_mem() free all memory used by nodes attributes
 */
void
i_free_mem()
{
	if (CurrentNode) {
		while (CurrentNode->parent)	/* make sure the CurrentNode is the root */
			CurrentNode = CurrentNode->parent;
		free_trees(CurrentNode);	/* call free_trees for the root tree */
	}
	CurrentNode = 0;				/* set the pointer to 0 to be able to create another graph in the graphical interface */
}

/*
 * static void free_attributes(struct attributes *attrs)
 * free the memory used by attrs
 */
static void
free_attributes(struct attributes* attrs)
{
	if (attrs) {
		free_attributes(attrs->next);
		free(attrs->a_name);
		free(attrs->value);
		free(attrs);
	}
}


/*
 * static void free_trees(struct node* the_tree) free all memory used by the_tree
 */
static void
free_trees(struct node* the_tree)
{
	if (the_tree->child) {
		struct node *tmp;
		free_trees(the_tree->child);
		tmp = the_tree->child->sibling;
		while (tmp) {
			free_trees(tmp);
			tmp = tmp->sibling;
		}
	}
	free_attributes(the_tree->syn_attributes);
	free_attributes(the_tree->inh_attributes);
	free(the_tree->label);
	free(the_tree);
}


/*
 * void i_batch_print(int argc, char **argv)
 * parse arguments for properties visual tree used only at the batch version
 */
void
i_init_dotty(int argc, char **argv)
{
	int option;			/* the current option */
	extern char *optarg;	/* the argument of the current option */
	extern int optind;		/* the option index */
	int rankdir = 1;		/* the default rankdir is TD */
	int depth = 0;			/* the default depth is the entiry tree */
	int termnr = 0;			/* the default root is the real ROOT of the tree */
	char *term = 0;
	char *attributes= 0;		/* the attributes to show, default all */
	char *fname = 0;		/* the output file name */
	int errflg = 0;			/* are there errors on the options? */
	char *options = "ld:t:ga:o:h";	/* the options... ':' means the option before has (at least) one argument */
	#ifdef WIN32
		int rundotty = 0;
	#endif

	while ((option = getopt(argc, argv, options)) != EOF) {		/* parse all options */
		switch (option) {
			case 'l':	/* change rankdir to lr */
				rankdir = 0;
				break;
			case 'd':	/* set the depth of the tree */
				depth = atoi(optarg);
				if (!depth) {
					fprintf(stderr,"Error: illegal input for depth %s\n.",optarg);
					errflg++;
				}
				break;
			case 't':	/* set the root of the tree */
				term = malloc(strlen(optarg));
				strcpy(term,optarg);
				if (argv[optind]) {
					termnr = atoi(argv[optind++]);
					if (!termnr) {
						fprintf(stderr,"Warning: illegal input for termnr %s.\n",argv[optind-1]);
						fprintf(stderr,"casting nr to 1.\n");
						termnr = 1;
					}
				} else {
					errflg++;
				}
				break;
			case 'g':	/* startup dotty after generating graph */
				#ifdef WIN32
					rundotty = 1;
				#else
					fprintf(stderr, "Warning: unknown option \"-g\" for this OS\n");
					fprintf(stderr, "Skipping ......\n");
				#endif
				break;
			case 'a':
				attributes = malloc(strlen(optarg));
				strcpy(attributes,optarg);
				break;
			case 'h':
				errflg++;
				break;
			case 'o':	/* set the output file */
				if (!strcmp(&optarg[strlen(optarg)-4],".dot")) {
					fname = malloc(strlen(optarg));
					strcpy(fname,optarg);
				} else {
					fname = malloc(strlen(optarg)+4);
					strcpy(fname,optarg);
					strcat(fname, ".dot");
				}
				break;
		}
	}
	for ( ; optind < argc; optind++) {
		fprintf(stderr, "Warning: unknown option %s\n", argv[optind]);
		errflg++;
	}
	if (errflg) {
		fprintf(stderr,"Usage: compiler inputfile [-a<a1,a2,...>-l-t<(non)terminal><occurance>-d<depth>-g-o<filename>]\n");
		fprintf(stderr,"\ta\tlist of attributes to show in the visible tree\n");
		fprintf(stderr,"\tl\tchange the print direction to Left-Right\n");
		fprintf(stderr,"\td\tthe depth of the visible tree\n");
		fprintf(stderr,"\tg\tstart dotty after generating the tree\n");
		fprintf(stderr,"\to\tprint the tree to the following dot-file (without extension)\n");
		fprintf(stderr,"\tt\tprint the tree from the following (non)terminal with occurance\n");
		exit (2);
	}

	if (!fname) {
		fname = malloc(strlen(argv[1])+4);
		strcpy(fname,argv[1]);
		strcat(fname, ".dot");
	}
	i_create_dot_file(fname, depth, rankdir, term, termnr,to_array(attributes));
	
	#ifdef WIN32
	if (rundotty)
		i_run_dotty(fname);
	#endif

	i_free_mem();
}


static char *
capitalize(char *string)
{
	char *returnvalue = malloc((strlen(string)+1));
	int n;
	for (n = 0; n < strlen(string); n++)
		returnvalue[n] = (char)toupper(string[n]);
		
	returnvalue[n]='\0';
	return returnvalue;
}

static int
in_attr_list(struct attributes* list, char* name)
{
	struct attributes* temp = list;
	while (temp) {
		if (!strcmp(temp->a_name,name))
			return 1;
		temp = temp->next;
	}
	return 0;
}

static char **
to_array(char* attrs)
{
	char** array = (char**)malloc(sizeof(char*)*(nr_of_attributes(attrs)+1));
	if (attrs) {
		int length = strcspn(attrs,",");
		int nr = 0;

		array[nr] = (malloc(length+1));
		strncpy(array[nr],attrs,length);
		array[nr][length] ='\0';
		length++;
		while (length <(strlen(attrs)+1)) {
			int templength = strcspn(&attrs[length],",");
			while (!templength)	/* skip double comma's */
				templength = strcspn(&attrs[++length],",");

			array[++nr] = (malloc(templength+1));
			strncpy(array[nr], &attrs[length], templength);
			array[nr][templength] ='\0';
			length = length+templength+1; /* skip the next comma */
		}
		array[nr+1] = 0;	/* set the last attribute to 0 */
		return array;
	} else
		array[0] = 0;
	return array;
}

static int
nr_of_attributes(char* attrs)
{
	int nr = 0;
	int count = 0;
	if (attrs) {
		for (; count < strlen(attrs); count++) {
			if (attrs[count] == ',')
				nr++;
		}
		return nr+1;
	} else
		return 0;
}
