/*******************************************************************************
* Implementation of the attribute evaluation actions, used in a SLADE
* parser specification
*
* JHG sep97 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/
#include "evalattr.h"

char *
i_boolean_text(boolean b)
{
	return b ? "true" : "false";
}

char *
i_char_text(char c)
{
	static char	text[4];

	sprintf(text, "'%c'", c);
	return text;
}

char *
i_integer_text(integer i)
{
	static char	text[10];

	sprintf(text, "%d", i);
	return text;
}

char *
i_real_text(real r)
{
	static char	text[20];

	sprintf(text, "%g", r);
	return text;
}

char *
i_string_text(string s)
{
	return s;
}

void assign_bool	(boolean *dest,		boolean src)	{ *dest= src; }
void assign_char	(char *dest,		char src)		{ *dest= src; }
void assign_int		(integer *dest,		integer src)	{ *dest= src; }
void assign_real	(real *dest,		real src)		{ *dest= src; }
void assign_string	(string *dest,		string src)		{ strcpy(*dest, src); }

void clear_bool		(boolean *dest)		{ *dest= false; }
void clear_char		(char *dest)		{ *dest= '\0'; }
void clear_int		(integer *dest)		{ *dest= 0; }
void clear_real		(real *dest)		{ *dest= 0.0; }
void clear_string	(string *dest)		{ (*dest)[0]= '\0'; }

void read_bool(boolean *arg, string prompt)
{
	char	answer[2];

	if (i_read_line(answer, 2, prompt) != 0)
		*arg= toupper(answer[0]) == 'T';
	else
		*arg= false;
}

void read_char(char *arg, string prompt)
{
	char	answer[2];

	if (i_read_line(answer, 2, prompt) != 0)
		*arg= *answer;
	else
		*arg= '\0';
}

void read_int(integer *arg, string prompt)
{
	char	answer[7];

	if (i_read_line(answer, 7, prompt) != 0)
		sscanf(answer, "%d", (int *) arg);
}

void read_real(real *arg, string prompt)
{
	char	answer[25];

	if (i_read_line(answer, 25, prompt) != 0)
		sscanf(answer, "%lg", arg);
}

void read_string(string *arg, string prompt)
{
	char	*s;

	if ((s= i_read_line(*arg, i_STRINGLEN+1, prompt)) != 0) {
		int		len= strlen(s);

		if (*(s+len-1) == '\n')
			*(s+len-1)= '\0';		/* remove trailing newline */
	} else
		(*arg)[0]= '\0';
}

void write_bool(boolean arg)
{
	i_report(i_TRACE, arg ? "true" : "false");
}

void write_char(char arg)
{
	i_report(i_TRACE, "%c"
	, (arg == '\r') ? '\n' : arg);	/* \r MS-DOS feature? :-( */
}

void write_int(integer arg)
{
	i_report(i_TRACE, "%d", arg);
}

void write_real(real arg)
{
	i_report(i_TRACE, "%g", arg);
}

void write_string(string arg)
{
	i_report(i_TRACE, arg);
}
