/*******************************************************************************
* Implementation of the run time system functions used in the code generated by
* the VIM to C translator
*
* JHG sep97 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "rts.h"

extern char		*StringTab[];

#define MAXRDSTRINGS	100	/* max. read string table size */
static int		RdStringTabSize;
static char		*RdStringTab[MAXRDSTRINGS];

#define MAXSTACK	100	/* max. stack size */
static int		Stack[MAXSTACK];

#define MAXDEPTH	15	/* max. nesting of blocks */
static int		CSG[MAXDEPTH];			/* segment group */
static int		Sp= 0, Csn= 0, Dsp= 0, *Data= NULL;

static jmp_buf		CReturnStack[MAXDEPTH];		/* v2c only */
static int		AsReturnStack[MAXDEPTH];	/* v2a only */
static int		Rsp= 0;

static void
fatal(char *s)
{
	fputs(s, stdout);
	exit(1);
}

int
pop(void)		/* v2c only, pop and return a value off the stack */
{
	if (Sp > 0)
		return Stack[--Sp];

	fatal("pop: empty stack\n");
	return 0;
}

int
push(int value)		/* v2c only, push value onto the stack and return it */
{
	if (Sp < MAXSTACK)
		Stack[Sp++]= value;
	else
		fatal("push: stack overflow\n");
	return value;
}

void
swap(void)		/* v2c only */
{
	int		i;

	i= Stack[Sp-1];
	Stack[Sp-1]= Stack[Sp-2];
	Stack[Sp-2]= i;
}

int
eqn(int len, int sn, int dpl)	/* v2c only */
{
	int		sp1, sp2;

	if (len == 0)
		len= getmemr(sn, dpl+2);

	if (Sp - 2*len < 0) {
		fatal("eqn: empty stack\n");
		return 0;
	}
	sp1= Sp-1;
	sp2= sp1-len;
	Sp -= 2*len;
	while (len--)
		if (Stack[sp1--] != Stack[sp2--])
			return 0;

	return 1;
}

void
crseg(int sn, int len)
{
#ifdef DEBUG
	fprintf(stdout, "crseg(sn=%d, len=%d)\n", sn, len);
#endif /* DEBUG */
	if (sn < 1 || sn > MAXDEPTH) {
		fatal("crseg: segment number exceeds maximum\n");
		return;
	}

	if (Data != NULL)
		Data= (int *)realloc(Data, (Dsp+len+4)*sizeof(int));
	else
		Data= (int *)malloc((Dsp+len+4)*sizeof(int));

	if (Data == NULL) {
		fatal("crseg: not enough memory\n");
		return;
	}

	Data[Dsp++]= sn;
	Data[Dsp++]= len;
	Data[Dsp++]= CSG[sn-1];			/* static */
	Data[Dsp++]= CSG[Csn];			/* dynamic */

	Csn= sn;
	CSG[Csn]= Dsp;
	Dsp += len;
}

void
dlseg(void)
{
	int	curseg, callseg, i, callsn, dynamic_link, static_link, len;
#ifdef DEBUG
	fprintf(stdout, "dlseg()\n");
#endif /* DEBUG */

	if ((curseg= CSG[Csn]) < 4) {
		fatal("dlseg: no current segment\n");
		return;
	}
	len= Data[curseg - 3];
	static_link= Data[curseg - 2];
	dynamic_link= Data[curseg - 1];
	if (static_link != dynamic_link) {
		callseg= dynamic_link;	/* caller is in different segment grp */
		callsn= Data[dynamic_link-4];/* restore segment grp of caller */
		for (i= callsn; i >= Csn; i--) {
			CSG[i]= callseg;		/* restore caller's environment */
			callseg= Data[callseg-2];
		}
		Csn= callsn;
	} else {
		if (Csn == 0) {
			fatal("dlseg: no more segments too delete\n");
			return;
		}
		CSG[Csn--]= 0;	/* same segment group */
	}
	Dsp -= len + 4;
	Data= (int *)realloc(Data, Dsp*sizeof(int));
}

int
getmemr(int sn, int dpl)
{
#ifdef DEBUG
	fprintf(stdout, "getmemr(sn=%d, dpl=%d) ", sn, dpl);
#endif /* DEBUG */
	if (sn > 0 && sn <= MAXDEPTH && CSG[sn] > 3 && CSG[sn] < Dsp) {
		if (dpl >= 0 && dpl < Data[CSG[sn]-3]) {
#ifdef DEBUG
	fprintf(stdout, "returns %d\n", Data[CSG[sn]+dpl]);
#endif /* DEBUG */
			return Data[CSG[sn]+dpl];
		}
		fatal("getmemr: wrong displacement\n");
	} else
		fatal("getmemr: wrong segment number\n");
	return 0;
}

void
putmemr(int sn, int dpl, int value)
{
#ifdef DEBUG
	fprintf(stdout, "putmemr(sn=%d, dpl=%d, value=%d)\n", sn, dpl, value);
#endif /* DEBUG */
	if (sn > 0 && sn <= MAXDEPTH && CSG[sn] > 3 && CSG[sn] < Dsp) {
		if (dpl >= 0 && dpl < Data[CSG[sn]-3]) {
			Data[CSG[sn]+dpl]= value;
			return;
		}
		fatal("putmemr: wrong displacement\n");
	} else
		fatal("putmemr: wrong segment number\n");
}

int
putxlowsiz(int sn, int dpl, int k, int size, int up, int low) /* for v2a only */
{
	size *= (up-low+1);
	putmemr(sn, dpl+2*k+1, low);
	putmemr(sn, dpl+2*k, size);
	return size;
}

int
getxlow(int sn, int dpl, int dim)			/* for v2a only */
{
	return getmemr(sn, dpl+2*dim+1);
}

int
getxsize(int sn, int dpl, int dim)			/* for v2a only */
{
	return getmemr(sn, dpl+2*dim);
}

int
getseglen(int sn)					/* also used by v2a */
{
#ifdef DEBUG
	fprintf(stdout, "getseglen(sn=%d) returns %d\n", sn, Data[CSG[sn]-3]);
#endif /* DEBUG */
	return Data[CSG[sn]-3];
}

void
incrseglen(int sn, int lenincr)				/* also used by v2a */
{
#ifdef DEBUG
	fprintf(stdout, "incrseglen(sn=%d, lenincr= %d)\n", sn, lenincr);
#endif /* DEBUG */
	Data[CSG[sn]-3] += lenincr;
	Dsp += lenincr;
	if ((Data= (int *)realloc(Data, Dsp*sizeof(int))) == NULL) {
		fatal("descr: not enough memory\n");
		return;
	}
}

void
descr(int sn, int dpl, int dim)				/* for v2c only */
{
	int	offset, size, k, up, lo;

#ifdef DEBUG
	fprintf(stdout, "descr(sn=%d, dpl= %d, dim=%d)\n", sn, dpl, dim);
#endif /* DEBUG */
	size= 1;
	for (k= dim; k > 0; k--) {
		up= pop();
		lo= pop();

		if (up < lo) {
			fatal("descr: upper bound smaller than lower bound\n");
			return;
		}

		size *= (up-lo+1);
		putmemr(sn, dpl+2*k+1, lo);
		putmemr(sn, dpl+2*k, size);
	}

	offset= getseglen(sn);
	putmemr(sn, dpl+1, offset);
	incrseglen(sn, size);
	putmemr(sn, dpl, dim);
}

int
ixdpl(int sn, int dpl)			/* for v2c only */
{
	int	idx, dim, low, size, nextsize, offset, xdpl;

#ifdef DEBUG
	fprintf(stdout, "ixdpl(sn=%d, dpl= %d) ", sn, dpl);
#endif /* DEBUG */
	dim= pop();
	size= 1;
	xdpl= 0;
	if (dim >= 0) {
		while (dim--) {
			idx= pop();
			low= pop();
			xdpl += (idx-low)*size;
			size= pop();
			if (idx < low || idx >= low+size) {
				fatal("index out of bounds\n");
				return -1;
			}
		}
		xdpl += dpl;
	} else {		/* if dim < 0 (dynamic array) get index from segment */
		dim= getmemr(sn, dpl);
		offset= getmemr(sn, dpl+1);
		while (dim) {
			idx= pop();
			low= getmemr(sn, dpl+2*dim+1);

			xdpl += (idx-low)*size;

			nextsize= getmemr(sn, dpl+2*dim);

			if ((idx-low)*size < 0 || (idx-low)*size > nextsize-size) {
				fatal("index out of bounds\n");
				return -1;
			}
			size= nextsize;
			dim--;
		}
		xdpl += offset;
	}
#ifdef DEBUG
	fprintf(stdout, "returns %d\n", xdpl);
#endif /* DEBUG */
	return xdpl;
}

int
varaddr(int sn, int dpl)
{
#ifdef DEBUG
	fprintf(stdout, "varaddr(sn=%d, dpl=%d) ", sn, dpl);
#endif /* DEBUG */
	if (sn > 0 && sn <= MAXDEPTH && CSG[sn] > 3 && CSG[sn] < Dsp) {
		if (dpl >= 0 && dpl < Data[CSG[sn]-3]) {
#ifdef DEBUG
			fprintf(stdout, "returns %d\n", CSG[sn]+dpl);
#endif /* DEBUG */
			return CSG[sn] + dpl;
		}
		fatal("varaddr: wrong displacement\n");
	} else
		fatal("varaddr: wrong segment number\n");
	return 0;
}

int
getmema(int address)
{
#ifdef DEBUG
	fprintf(stdout, "getmema(address=%d) ", address);
#endif /* DEBUG */
	if (address > 3 && address <= Dsp) {
#ifdef DEBUG
		fprintf(stdout, "returns %d\n", Data[address]);
#endif /* DEBUG */
		return Data[address];
	}
	fatal("getmema: illegal absolute address\n");
	return 0;
}

void
putmema(int value, int address)
{
#ifdef DEBUG
	fprintf(stdout, "putmema(address=%d, value=%d)\n", address, value);
#endif /* DEBUG */
	if (address > 3 && address <= Dsp)
		Data[address]= value;
	else
		fatal("putmema: illegal absolute address\n");
}

void
ccall(jmp_buf *jb)			/* v2c only */
{
#ifdef DEBUG
	fprintf(stdout, "call(jmp_buf*= %u) Rsp= %d\n", (unsigned)jb, Rsp);
#endif /* DEBUG */
	/* pushes context jb onto return stack */
	if (Rsp < MAXDEPTH)
		memcpy(&CReturnStack[Rsp++], jb, sizeof(jmp_buf));
	else
		fatal("call: return stack overflow\n");
}

void
creturn_(jmp_buf *jb)			/* v2c only */
{
#ifdef DEBUG
	fprintf(stdout, "return_(jmp_buf*= %u), Rsp=%d\n", (unsigned)jb, Rsp-1);
#endif /* DEBUG */
	/* pops the top context from return stack and stores it in *jb */
	if (Rsp > 0)
		memcpy(jb, &CReturnStack[--Rsp], sizeof(jmp_buf));
	else
		fatal("return_: empty return stack\n");
}

void
ascall(int retaddr)			/* v2a only */
{
	/* pushes address retaddr onto return address stack */
#ifdef DEBUG
	fprintf(stdout, "call(retaddr= 0x%x) Rsp= %d\n", retaddr, Rsp);
#endif /* DEBUG */
	if (Rsp < MAXDEPTH)
		AsReturnStack[Rsp++]= retaddr;
	else
		fatal("call: return stack overflow\n");
}

int
asreturn_(void)				/* v2a only */
{
	/* returns return address popped off top of return address stack */
#ifdef DEBUG
	fprintf(stdout, "return_(), Rsp=%d ", Rsp-1);
#endif /* DEBUG */
	if (Rsp > 0) {
#ifdef DEBUG
		fprintf(stdout, "returns 0x%x\n", AsReturnStack[Rsp-1]);
#endif /* DEBUG */
		return AsReturnStack[--Rsp];
	} else
		fatal("return_: empty return stack\n");
	return 0;
}

void
wrbool(int value)
{
	fputs(value ? "true\n" : "false\n", stdout);
}

void
wrchar(int value)
{
	fprintf(stdout, "%c\n", value);
}

void
wrint(int value)
{
	fprintf(stdout, "%d\n", value);
}

void
wrstring(int tabsize, int index)
{
	char	*s, *error;

#ifdef DEBUG
	fprintf(stdout, "wrstring(index=%d, tabsize=%d)\n", index, tabsize);
#endif /* DEBUG */
	error= "wrstring: unknown string??\n";
	if (index < 1 || index > tabsize + RdStringTabSize) {
		fputs(error, stdout);
		return;
	}
	if ((index > tabsize && (s= RdStringTab[index-tabsize-1]) != NULL)
	|| (s= StringTab[index-1]) != NULL)
		fprintf(stdout, "%s\n", s);
	else
		fputs(error, stdout);
}

int
rdbool(void)
{
	char	inputbuf[8];

	fgets(inputbuf, 7, stdin);
	return (inputbuf[0] == 't' || inputbuf[0] == 'T');
}

int
rdchar(void)
{
	char	inputbuf[4];

	fgets(inputbuf, 3, stdin);
	return inputbuf[0];
}

int
rdint(void)
{
	char	inputbuf[8];

	fgets(inputbuf, 7, stdin);
	return atoi(inputbuf);
}

int
rdstring(int tabsize)
{
	char	inputbuf[256];

	fgets(inputbuf, 255, stdin);
	inputbuf[strlen(inputbuf)-1]= 0;		/* remove trailing newline */
	if (RdStringTabSize < MAXRDSTRINGS) {
		RdStringTab[RdStringTabSize++]= strdup(inputbuf);
		return RdStringTabSize + tabsize;
	}
	fatal("rdstring: string table full\n");
	return 0;
}
