/*******************************************************************************
* Implementation of the run time system functions used in the code generated by
* the VIM to JVM translator
*
* JHG nov98 SLADE Copyright (C) 1991-1997 University of Twente
*******************************************************************************/

import java.io.*;
import java.util.*;


public class rts {
	final private int		MAXDEPTH= 15; 	/* max. nesting of blocks */

	private int				CSG[]= new int[MAXDEPTH];		/* segment group */
	private int				Csn= 0, Dsp= 0, Rsp= 0;
	private Vector			Data= new Vector();
	private int				ReturnStack[]= new int[MAXDEPTH];

	static BufferedReader	stdin= new BufferedReader(
								new InputStreamReader(System.in)
							);
	static PrintStream		stdout= System.out;
	private boolean			Debug= false;

	private void fatal(String s)
	{
		System.out.print(s);
		System.exit(1);
	}

	private void setData(int value, int index)
	{
		Data.setElementAt(new Integer(value), index);
	}

	private int getData(int index)
	{
		if (Data.elementAt(index) != null)
			return ((Integer)(Data.elementAt(index))).intValue();
		return 0;
	}

	public void crseg(int sn, int len)
	{
		if (Debug)
			System.out.println("crseg(" + sn + ", " + len + ");");

		if (sn < 1 || sn > MAXDEPTH) {
			fatal("crseg: segment number exceeds maximum\n");
			return;
		}

		Data.setSize(Dsp+len+4);
		setData(sn, Dsp++);
		setData(len, Dsp++);
		setData(CSG[sn-1], Dsp++);		/* static */
		setData(CSG[Csn], Dsp++);		/* dynamic */

		Csn= sn;
		CSG[Csn]= Dsp;
		Dsp += len;
	}

	public void dlseg()
	{
		int	curseg, callseg, i, callsn, dynamic_link, static_link, len;

		if (Debug)
			System.out.println("dlseg()");

		if ((curseg= CSG[Csn]) < 4) {
			fatal("dlseg: no current segment\n");
			return;
		}
		len= getData(curseg - 3);
		static_link= getData(curseg - 2);
		dynamic_link= getData(curseg - 1);
		if (static_link != dynamic_link) {
			callseg= dynamic_link;	/* caller is in different segment grp */
			callsn= getData(dynamic_link-4); /* restore segment grp of caller */

			for (i= callsn; i >= Csn; i--) {
				CSG[i]= callseg;		/* restore caller's environment */
				callseg= getData(callseg-2);
			}
			Csn= callsn;
		} else {
			if (Csn == 0) {
				fatal("dlseg: no more segments too delete\n");
				return;
			}
			CSG[Csn--]= 0;	/* same segment group */
		}
		Dsp -= len + 4;
		Data.setSize(Dsp);
	}

	public int getmemr(int sn, int dpl)
	{
		if (Debug)
			System.out.println("getmemr(" + sn + ", " + dpl + ");");

		if (sn > 0 && sn <= MAXDEPTH && CSG[sn] > 3 && CSG[sn] < Dsp) {
			if (dpl >= 0 && dpl < getData(CSG[sn]-3)) {
				return getData(CSG[sn]+dpl);
			}
			fatal("getmemr: wrong displacement\n");
		} else
			fatal("getmemr: wrong segment number\n");
		return 0;
	}

	public void putmemr(int value, int sn, int dpl)
	{
		if (Debug)
			System.out.println("putmemr(" + value +", "+ sn +", "+ dpl +");");

		if (sn > 0 && sn <= MAXDEPTH && CSG[sn] > 3 && CSG[sn] < Dsp) {
			if (dpl >= 0 && dpl < getData(CSG[sn]-3)) {
				setData(value, CSG[sn]+dpl);
				return;
			}
			fatal("putmemr: wrong displacement\n");
		} else
			fatal("putmemr: wrong segment number\n");
	}

	public int varaddr(int sn, int dpl)
	{
		if (Debug)
			System.out.println("varaddr(" + sn + ", " + dpl + ");");

		if (sn > 0 && sn <= MAXDEPTH && CSG[sn] > 3 && CSG[sn] < Dsp) {
			if (dpl >= 0 && dpl < getData(CSG[sn]-3)) {
				return CSG[sn] + dpl;
			}
			fatal("varaddr: wrong displacement\n");
		} else
			fatal("varaddr: wrong segment number\n");
		return 0;
	}

	public int getmema(int address)
	{
		if (Debug)
			System.out.println("getmema(" + address + ");");

		if (address > 3 && address <= Dsp) {
			return getData(address);
		}
		fatal("getmema: illegal absolute address\n");
		return 0;
	}

	public void putmema(int value, int address)
	{
		if (Debug)
			System.out.println("putmema(" + value + ", " + address + ");");

		if (address > 3 && address <= Dsp)
			setData(value, address);
		else
			fatal("putmema: illegal absolute address\n");
	}

	public int getseglen(int sn)
	{
		return getData(CSG[sn]-3);
	}

	public void incrseglen(int sn, int lenincr)
	{
		setData(getData(CSG[sn]-3) + lenincr, CSG[sn]-3);
		Dsp += lenincr;
		Data.setSize(Dsp);
	}

	public void pushReturn(int retaddr)
	{
		/* pushes address retaddr onto return address stack */
		if (Rsp < MAXDEPTH)
			ReturnStack[Rsp++]= retaddr;
		else
			fatal("call: return stack overflow\n");
	}

	public int popReturn()
	{
		/* returns return address popped off top of return address stack */
		if (Rsp > 0)
			return ReturnStack[--Rsp];
		fatal("return_: empty return stack\n");
		return 0;
	}
}
